package Mailtrace::Daemon::Parsing;

# Includes
use strict;
use warnings;

use Data::Dumper;
use Mailtrace::Database::Layer;

# Constructor
sub new {
	my ($invocant, $basic_functions, $logging, $configuration, $memd, $cache_expire_time) = @_;
	
	# Was the constructor called by an instance or an class?
	my $class = ref($invocant) || $invocant;
	
	# Precompile the regexp
        my @reg_exp = map { qr/$_/x } (
		'^(.+)\s(.+)\spostfix.+NOQUEUE:\s+reject:\s+RCPT\s+from\s+(.+)\[(.+)\]:\s+(.+);\s+from=<([^\s]+)@([^\s]+)>\s+to=<([^\s]+)@([^\s]+)>\s+proto.+',
		'^.+\s.+\samavis.+:\s\((.+)\)\s+SPAM,\s+<[^,]+>\s+->\s+<[^,]+>,\s+Yes,\s+score=.+\s+tag=.+\s+tests=\[(.+)\]\s+autolearn=.+',
		'^(.+)\s(.+)\samavis.+:\s\((.+)\)\s+Blocked\s+SPAM,\s+\[.+\]\s+\[(.+)\]\s+<([^\s]+)@([^\s]+)>\s+->\s+<([^\s]+)@([^\s]+)>,\s+Message-ID:\s+<(.+)>,\s+mail_id:\s+.+,\s+Hits:\s+(.+),\s+size:\s+(.+),\s+.+',
		'^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\sclient=(.+)\[(.+)\]:([[:xdigit:]]+),\sorig_client=(.+)\[(.+)\]:.*',
		'^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\sclient=(.+)\[(.+)\]',
		':\s+([[:xdigit:]]+):\s+message-id=<(.+)>',
		'^.+\spostfix.+:\s([[:xdigit:]]+):\s+from=<(?:([^\s]+)@([^\s]+))*>,\s+size=(.+),.+',
		'^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+conn_use=[[:xdigit:]]+,\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)',
		'^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)',
		'.+said:\s5\d\d(?:\s|-).+',
		'^(.+)\s(.+)\spostfix.+:\s+([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+orig_to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+conn_use=[[:xdigit:]]+,\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)',
		'^(.+)\s(.+)\spostfix.+:\s+([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+orig_to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)',
		'connect\s+to\s+(.+\[.+\]):.+',
		# This is the regexp exclude for tuning the matching routine
		'(?:\]:\s450\s4\.)|(?:\]:\sconnect\sfrom)|(?:\]:\slost\sconnection\safter)|(?:action=PREPEND)|(?:\]:\sweighted\scheck:)|(?:[[:xdigit:]]+:\sremoved)|(?:\]:\sdisconnect\sfrom)'
	);

	# Generate the month table for faster date conversion
	my %month_table = (
	   Jan => '01',
	   Feb => '02',
	   Mar => '03',
	   Apr => '04',
	   May => '05',
	   Jun => '06',
	   Jul => '07',
	   Aug => '08',
	   Sep => '09',
	   Oct => '10',
	   Nov => '11',
	   Dec => '12'
	);
	
	# Instance is a hash reference
	my $self = {
        configuration     => $configuration,
        basic_functions   => $basic_functions,
        logging           => $logging,
        regex             => \@reg_exp,
        database          => 'NONE',
        memcached         => $memd,
        cache_expire_time => $cache_expire_time,
        year              => qx(date +%Y),
        month_table       => \%month_table,
        mydomain          => 'NONE'
	};
	
	# Make hash reference to an instance
	bless ($self, $class);
	
	# get mydomain
	$self->{mydomain} = $self->get_mydomain('foo');
	
	# Clean linebreaks in the year string
	$self->{year} =~ s/\s+//x;
	
	# Create the database instance
	$self->{database} = Mailtrace::Database::Layer->new(
	    $self->{basic_functions},
	    $self->{logging},
	    $self->{configuration},
	    $self->{memcached}
	) or do {
		$self->{basic_functions}->alert(0, "Unable to connect the database!");
	};
	
	# Return the instance
    return $self
}

# Destructor
sub DESTROY {
	my ($self) = @_;
	
	$self->{logging}->debug(1, "Destroying analyzation object");
	
	# Check for an overridden destructor...
    $self->SUPER::DESTROY if $self->can("SUPER::DESTROY");
    
    return;
}

# Gets mydomain (if not defined in config, run postconf mydomain. return error if postconf doesnt exist)
sub get_mydomain {
    my ($self, $foo) = @_;
    
    my $mydomain = '';
    if (exists $self->{configuration}{general}{mydomain}){
        $mydomain = $self->{configuration}{general}{mydomain}
    } else {
        my $postconf_exists = `which -a postconf`;
        if ($postconf_exists ne ''){
            my $postconf_answer = `postconf mydomain`;
            if ( $postconf_answer eq ''){
                # $self->{basic_functions}->alert(1, "postconf failed: $!\n");
                $self->{logging}->debug(1, "postconf failed: $!\n");
            } else {
                $mydomain = substr($postconf_answer, 11);
            }
        } else {
            $self->{logging}->debug(1, "postconf not found!\n");
        }
    }
    return \$mydomain;
}

# Analyze the given line
sub analyze_line {
	my ($self, $log_line) = @_;
	
	# Skip all, if we definitive have an uninteresting line here
	return if $$log_line =~ /${$self->{regex}}[10]/x;
        
	# Look for blocked deliveries after RCPT (Single-entry, no memcache needed)
	# ^(.+)\s(.+)\spostfix.+NOQUEUE:\s+reject:\s+RCPT\s+from\s+(.+)\[(.+)\]:\s+(5.+);\s+from=<([^\s]+)@([^\s]+)>\s+to=<([^\s]+)@([^\s]+)>\s+proto.+
    if ($$log_line =~ /${$self->{regex}}[0]/x) {
    	$self->{logging}->debug(1, "Regex found: 'blocked deliveries' at Line " . $$log_line);
        $self->process_blocked_entry({
            time             => $1,
            server           => $2,
            client_hostname  => lc($3),
            client_ip        => lc($4),
            reason           => $5,
            mailfrom_address => lc($6),
            mailfrom_domain  => lc($7),
            mailto_address   => lc($8),
            mailto_domain    => lc($9)
        });
        
	# Find amavis-Line with sa-scoring
	# ^.+\s.+\samavis.+:\s\((.+)\)\s+SPAM,\s+<[^,]+>\s+->\s+<[^,]+>,\s+Yes,\s+score=.+\s+tag=.+\s+tests=\[(.+)\]\s+autolearn=.+
    } elsif ($$log_line =~ /${$self->{regex}}[1]/x) {
    	$self->{logging}->debug(1, "Regex found: 'amavis' at Line " . $$log_line);
        $self->process_amavis_sa_scoring_entry({
            queueid  => $1,
            saalerts => $2
        });
	
	# Look for Amavis "Blocked Spam" (Memcache Entry will be deleted, because it is not needed anymore)
	# ^(.+)\s(.+)\samavis.+:\s\((.+)\)\s+Blocked\s+SPAM,\s+\[.+\]\s+\[(.+)\]\s+<([^\s]+)@([^\s]+)>\s+->\s+<([^\s]+)@([^\s]+)>,\s+Message-ID:\s+<(.+)>,\s+mail_id:\s+.+,\s+Hits:\s+(.+),\s+size:\s+(.+),\s+.+
    } elsif ($$log_line =~ /${$self->{regex}}[2]/x) {
    	$self->{logging}->debug(1, "Regex found: 'Amavis Blocked Spam' at Line " . $$log_line);    
        $self->process_amavis_blocked_spam({
            time             => $1,
            server           => $2,
            queueid          => $3,
            client_ip        => $4,
            mailfrom_address => lc($5),
            mailfrom_domain  => lc($6),
            mailto_address   => lc($7),
            mailto_domain    => lc($8),
            messageid        => $9,
            score            => $10,
            size             => $11
        });
        
	# Find a postfix connect line with orig_client
	# ^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\sclient=(.+)\[(.+)\]:(::xdigit:]]+),\sorig_client=(.+)\[(.+)\]:.*'.*  FIXME
    } elsif ($$log_line =~ /${$self->{regex}}[3]/x) {
    	$self->{logging}->debug(1, "Regex found: 'postfix connect line with orig_client' at Line " . $$log_line);
        $self->process_postfix_connect_entry({
            time            => $1,
            server          => $2,
            queueid         => $3,
            client_hostname => $7,
            client_ip       => $8
        });
        
	# Find a postfix connect line
	# ^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\sclient=(.+)\[(.+)\].*
    } elsif ($$log_line =~ /${$self->{regex}}[4]/x) {
    	$self->{logging}->debug(1, "Regex found: 'postfix connect line' at Line " . $$log_line);
        $self->process_postfix_connect_entry({
            time            => $1,
            server          => $2,
            queueid         => $3,
            client_hostname => $4,
            client_ip       => $5
        });
        
	# Find a Message-ID line
	# :\s+([[:xdigit:]]+):\s+message-id=<(.+)>
    } elsif ($$log_line =~ /${$self->{regex}}[5]/x) {
    	$self->{logging}->debug(1, "Regex found: 'Message-ID line' at Line " . $$log_line);
        $self->process_postfix_messageid_entry({
           queueid   => $1,
           messageid => $2
        });
        
	# Find a FROM line
	# ^.+\spostfix.+:\s([[:xdigit:]]+):\s+from=<(?:([^\s]+)@([^\s]+))*>,\s+size=(.+),.+
    } elsif ($$log_line =~ /${$self->{regex}}[6]/x) {
    	$self->{logging}->debug(1, "Regex found: 'FROM line' at Line " . $$log_line);
        my ($mailfrom_address, $mailfrom_domain);
        
        # Check if a MAILFROM is given, else set MAILER-DAEMON
        if (defined $2) {
            $mailfrom_address = lc($2);
            $mailfrom_domain  = lc($3);
        } else {
            $mailfrom_address = "MAILER-DAEMON";
            $mailfrom_domain  = "NONE";
        }
        
        $self->process_postfix_from_entry({
            queueid          => $1,
            mailfrom_address => $mailfrom_address,
            mailfrom_domain  => $mailfrom_domain,
            size             => $4
        });
            
    # Find a delivery line without orig_to but with conn_use
	# ^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),,\s+conn_use=[:xdigit:],\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)
    } elsif ($$log_line =~ /${$self->{regex}}[7]/x) {
    	$self->{logging}->debug(1, "Regex found: 'delivery line without orig_to' at Line " . $$log_line);
        $self->process_postfix_no_origto_entry({
            time           => $1,
            server         => $2,
            queueid        => $3,
            mailto_address => lc($4),
            mailto_domain  => lc($5),
            relayhost      => lc($6),
            status         => lc($7),
            reason         => $8
        });
        
        # Find a delivery line without orig_to and without conn_use
	# ^(.+)\s(.+)\spostfix.+:\s([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)
    } elsif ($$log_line =~ /${$self->{regex}}[8]/x) {
    	$self->{logging}->debug(1, "Regex found: 'delivery line with orig_to' at Line " . $$log_line);
        $self->process_postfix_no_origto_entry({
            time           => $1,
            server         => $2,
            queueid        => $3,
            mailto_address => lc($4),
            mailto_domain  => lc($5),
            relayhost      => lc($6),
            status         => lc($7),
            reason         => $8
        });
        
	# Find a delivery line with orig_to and conn_use
	# '^(.+)\s(.+)\spostfix.+:\s+([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+orig_to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+conn_use=[:xdigit:],\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)'
    } elsif ($$log_line =~ /${$self->{regex}}[10]/x) {
    	$self->{logging}->debug(1, "Regex found: 'delivery line with orig_to and conn_use' at Line " . $$log_line);
        $self->process_postfix_origto_entry({
            time                => $1,
            server              => $2,
            queueid             => $3,
            mailto_address      => lc($4),
            mailto_domain       => lc($5),
            original_to_address => lc($6),
            original_to_domain  => lc($7),
            relayhost           => lc($8),
            status              => lc($9),
            reason              => $10
        });

	# Find a delivery line with orig_to but without conn_use
	# ^(.+)\s(.+)\spostfix.+:\s+([[:xdigit:]]+):\s+to=<([^\s]+)@([^\s]+)>,\s+orig_to=<([^\s]+)@([^\s]+)>,\s+relay=(.+),\s+delay=.+,\s+status=([^\s]+)\s+\((.+)\)
    } elsif ($$log_line =~ /${$self->{regex}}[11]/x) {
    	$self->{logging}->debug(1, "Regex found: 'delivery line with orig_to but without conn_use' at Line " . $$log_line);
        $self->process_postfix_origto_entry({
            time                => $1,
            server              => $2,
            queueid             => $3,
            mailto_address      => lc($4),
            mailto_domain       => lc($5),
            original_to_address => lc($6),
            original_to_domain  => lc($7),
            relayhost           => lc($8),
            status              => lc($9),
            reason              => $10
        });

	}
	
	return;
}

sub process_blocked_entry {
    my ($self, $arg_ref) = @_;
    my $DBID = undef;
    my $table_timestamp;
    
    # Do a server name mapping (if not defined in config leave the name as it is)
    $arg_ref->{server} = ${$self->{configuration}}{general}{myhostname}
        if defined ${$self->{configuration}}{general}{myhostname};
    
    # Convert the time
    ($arg_ref->{time}, $table_timestamp) = $self->convert_time($arg_ref->{time});
	
    # Save the dataset into connectlog database
    $DBID = $self->{database}->set_connectlog({
        table_timestamp  => $table_timestamp,
        client_hostname  => $arg_ref->{client_hostname},
        client_ip        => $arg_ref->{client_ip},
        mailfrom_address => $arg_ref->{mailfrom_address},
        mailfrom_domain  => $arg_ref->{mailfrom_domain},
        messageid        => "unknown",
        size             => "0",
        queueid          => "NOQUEUE",
        server           => $arg_ref->{server},
        time             => $arg_ref->{time}
    });
    
    # Save the extracted values to deliverlog database
    $self->save_deliver_entry({
        table_timestamp     => $table_timestamp,
        time                => $arg_ref->{time},
        server              => $arg_ref->{server},
        queueid             => 'NOQUEUE',
        mailto_address      => $arg_ref->{mailto_address},
        mailto_domain       => $arg_ref->{mailto_domain},
        relayhost           => 'invalid',
        status              => 'rejected',
        reason              => $arg_ref->{reason},
        dbid                => $DBID
    });
    
    return;
}

sub process_amavis_sa_scoring_entry {
    my ($self, $arg_ref) = @_;
    my %clCollector;
		
    # Get an possible existing incomplete dataset (normally this couldn't be)
    if (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
        %clCollector = %{${$self->{memcached}}->get($arg_ref->{queueid})};
        ${$self->{memcached}}->delete($arg_ref->{queueid});
    }
		
    # Save the found value
    $clCollector{SAALERTS} = $arg_ref->{saalerts};
		
    # Save the dataset back into memcached
    ${$self->{memcached}}->set($arg_ref->{queueid}, \%clCollector, $self->{cache_expire_time});
    
    return;
}

sub process_amavis_blocked_spam {
    my ($self, $arg_ref) = @_;
    my $relayhost = "invalid";
    my $status    = "rejected";
    my $reason    = "Die Nachricht wurde von Amavis/SpamAssassin als Spam bewertet und darum geblockt.";
    my $DBID      = undef;
    my %clCollector;
    my $table_timestamp;
    
    # Get an possible existing incomplete dataset
    if (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
        %clCollector = %{${$self->{memcached}}->get($arg_ref->{queueid})};
        ${$self->{memcached}}->delete($arg_ref->{queueid});
    }
    
    # If an scoring and test results exist, attach them to the reason string
    if (defined $clCollector{SAALERTS}) {
        $reason .= " Score: ".$arg_ref->{score}." Alerts: ".$clCollector{SAALERTS};
    }
    
    # Do a server name mapping (if not defined in config leave the name as it is)
    $arg_ref->{server} = ${$self->{configuration}}{general}{myhostname}
        if defined ${$self->{configuration}}{general}{myhostname};
    
    # Convert the time
    ($arg_ref->{time}, $table_timestamp) = $self->convert_time($arg_ref->{time});
	
    # Save the dataset into connectlog database
    $DBID = $self->{database}->set_connectlog({
        table_timestamp  => $table_timestamp,
        client_hostname  => 'unresolved',
        client_ip        => $arg_ref->{client_ip},
        mailfrom_address => $arg_ref->{mailfrom_address},
        mailfrom_domain  => $arg_ref->{mailfrom_domain},
        messageid        => $arg_ref->{messageid},
        size             => $arg_ref->{size},
        queueid          => $arg_ref->{queueid},
        server           => $arg_ref->{server},
        time             => $arg_ref->{time}
    });
	
    # Save the extracted values to deliverlog database
    $self->save_deliver_entry({
        table_timestamp     => $table_timestamp,
        time                => $arg_ref->{time},
        server              => $arg_ref->{server},
        queueid             => $arg_ref->{queueid},
        mailto_address      => $arg_ref->{mailto_address},
        mailto_domain       => $arg_ref->{mailto_domain},
        relayhost           => $relayhost,
        status              => $status,
        reason              => $reason,
        dbid                => $DBID
    });
    
    return;
}

sub process_postfix_connect_entry {
    my ($self, $arg_ref) = @_;
    my %connectlog_collector;
    my $table_timestamp;
    
    # Get an possible existing incomplete dataset
    if (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
        %connectlog_collector = %{${$self->{memcached}}->get($arg_ref->{queueid})};
        ${$self->{memcached}}->delete($arg_ref->{queueid});
    }
    
    # Convert the time
    ($arg_ref->{time}, $table_timestamp) = $self->convert_time($arg_ref->{time});
    
    # Save the found values temporary for later processing
    $connectlog_collector{CLIENT_HOSTNAME} = $arg_ref->{client_hostname};
    $connectlog_collector{CLIENT_IP}       = $arg_ref->{client_ip};
    $connectlog_collector{SERVER}          = $arg_ref->{server};
    $connectlog_collector{TIME}            = $arg_ref->{time};
    $connectlog_collector{TABLETIMESTAMP}  = $table_timestamp;
    
    # Save the dataset back into memcached
    ${$self->{memcached}}->set(
        $arg_ref->{queueid},
        \%connectlog_collector,
        $self->{cache_expire_time}
    );
    
    return 1;
}

sub process_postfix_messageid_entry {
    my ($self, $arg_ref) = @_;
    my %connectlog_collector;
    
    # Get an possible existing incomplete dataset
    if (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
        %connectlog_collector = %{${$self->{memcached}}->get($arg_ref->{queueid})};
        ${$self->{memcached}}->delete($arg_ref->{queueid});
    }
    
    # Save the found values temporary for later processing
    $connectlog_collector{MESSAGEID} = $arg_ref->{messageid};
        
    # Save the dataset back into memcached
    ${$self->{memcached}}->set(
        $arg_ref->{queueid},
        \%connectlog_collector,
        $self->{cache_expire_time}
    );
    
    return;
}

sub process_postfix_from_entry {
    my ($self, $arg_ref) = @_;
    my %connectlog_collector;
    
    # Get an possible existing incomplete dataset
    if (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
        %connectlog_collector = %{${$self->{memcached}}->get($arg_ref->{queueid})};
        ${$self->{memcached}}->delete($arg_ref->{queueid});
    }
    
    # Save the found values temporary for later processing
    $connectlog_collector{MAILFROM_ADDRESS} = $arg_ref->{mailfrom_address};
    $connectlog_collector{MAILFROM_DOMAIN} = $arg_ref->{mailfrom_domain};
    $connectlog_collector{SIZE} = $arg_ref->{size};
    
    # Check the completeness of the dataset and check if we should save deliverable/undeliverable entries
    if ($self->check_completeness(\%connectlog_collector)) {
	# If not explicitly set to ignore save it
	if (lc(${$self->{configuration}}{general}{ignore_verify_log_entries}) ne 'yes') {
	    my $DBID = undef;

	    # Do a server name mapping (if not defined in config leave the name as it is)
            $connectlog_collector{SERVER} = ${$self->{configuration}}{general}{myhostname}
                if defined ${$self->{configuration}}{general}{myhostname};

            $DBID = $self->{database}->set_connectlog({
                table_timestamp  => $connectlog_collector{TABLETIMESTAMP},
                client_hostname  => $connectlog_collector{CLIENT_HOSTNAME},
                client_ip        => $connectlog_collector{CLIENT_IP},
                mailfrom_address => $arg_ref->{mailfrom_address},
                mailfrom_domain  => $arg_ref->{mailfrom_domain},
                messageid        => $connectlog_collector{MESSAGEID},
                size             => $arg_ref->{size},
                queueid          => $arg_ref->{queueid},
                server           => $connectlog_collector{SERVER},
                time             => $connectlog_collector{TIME}
            });

            # Update the dataset with the DBID and the table_timestamp
            $connectlog_collector{DBID} = $DBID if $DBID > 0;

            # And put the dataset to memcache
            ${$self->{memcached}}->set($arg_ref->{queueid}, \%connectlog_collector, $self->{cache_expire_time});



	} else {
#	    if (!(($arg_ref->{mailfrom_address} eq ${$self->{configuration}}{general}{ignore_connects_from}) && ($arg_ref->{mailfrom_domain} =~ /^$connectlog_collector{SERVER}.*/))) {
            if (!(($arg_ref->{mailfrom_address} eq ${$self->{configuration}}{general}{ignore_connects_from}) && ($arg_ref->{mailfrom_domain} =~ /^(.*\.?$self->{my_domain}|localhost)/))) {
		my $DBID = undef;

                # Do a server name mapping (if not defined in config leave the name as it is)
                $connectlog_collector{SERVER} = ${$self->{configuration}}{general}{myhostname}
                    if defined ${$self->{configuration}}{general}{myhostname};
 
                $DBID = $self->{database}->set_connectlog({
                    table_timestamp  => $connectlog_collector{TABLETIMESTAMP},
                    client_hostname  => $connectlog_collector{CLIENT_HOSTNAME},
                    client_ip        => $connectlog_collector{CLIENT_IP},
                    mailfrom_address => $arg_ref->{mailfrom_address},
                    mailfrom_domain  => $arg_ref->{mailfrom_domain},
                    messageid        => $connectlog_collector{MESSAGEID},
                    size             => $arg_ref->{size},
                    queueid          => $arg_ref->{queueid},
                    server           => $connectlog_collector{SERVER},
                    time             => $connectlog_collector{TIME}
                });

                # Update the dataset with the DBID and the table_timestamp
                $connectlog_collector{DBID} = $DBID if $DBID > 0;

                # And put the dataset to memcache
                ${$self->{memcached}}->set($arg_ref->{queueid}, \%connectlog_collector, $self->{cache_expire_time});
	    }
	}
        
    # TODO: Other parts need an completeness check as well
    # If not complete, save all to the cache
    } else {
        ${$self->{memcached}}->set($arg_ref->{queueid}, \%connectlog_collector, $self->{cache_expire_time});
    }
    
    return;
}

sub process_postfix_no_origto_entry {
    my ($self, $arg_ref) = @_;
    my $table_timestamp;
    
    # Convert the time
    ($arg_ref->{time}, $table_timestamp) = $self->convert_time($arg_ref->{time});
    
    # Check if the mail was rejected or not
    # '.*said:\s5\d\d(?:\s|-).*'
    if (($arg_ref->{status} eq "bounced") && ($arg_ref->{reason} =~ /${$self->{regex}}[7]/x)) {
        $arg_ref->{status} = "rejected";
    }
    
    # Spezialitaet von Postfix: Host unreachable liefert "relay=none". Host muss aus SMTP-Error ausgeschnitten werden
    # connect\s+to\s+(.*\[.*\]):.*
    if ($arg_ref->{relayhost} eq "none") {
        # 'connect to (.*)\[.*\].*'
        $arg_ref->{reason} =~ /${$self->{regex}}[9]/x;
        $arg_ref->{relayhost} = lc($1);
    }
    # Save the extracted values to database
# If not explicitly set to ignore save it
    if (lc(${$self->{configuration}}{general}{ignore_verify_log_entries} ne 'yes')) {
	$self->save_deliver_entry({
            table_timestamp     => $table_timestamp,
            time                => $arg_ref->{time},
            server              => $arg_ref->{server},
            queueid             => $arg_ref->{queueid},
            mailto_address      => $arg_ref->{mailto_address},
            mailto_domain       => $arg_ref->{mailto_domain},
            relayhost           => $arg_ref->{relayhost},
            status              => $arg_ref->{status},
            reason              => $arg_ref->{reason},
        });
    } else {
	if (!(($arg_ref->{status} eq 'deliverable') || ($arg_ref->{status} eq 'undeliverable'))) {
	    $self->save_deliver_entry({
                table_timestamp     => $table_timestamp,
                time                => $arg_ref->{time},
                server              => $arg_ref->{server},
                queueid             => $arg_ref->{queueid},
                mailto_address      => $arg_ref->{mailto_address},
                mailto_domain       => $arg_ref->{mailto_domain},
                relayhost           => $arg_ref->{relayhost},
                status              => $arg_ref->{status},
                reason              => $arg_ref->{reason},
            });
        }
    }

    return;
}

sub process_postfix_origto_entry {
    my ($self, $arg_ref) = @_;
    my $table_timestamp;
    
    # Convert the time
    ($arg_ref->{time}, $table_timestamp) = $self->convert_time($arg_ref->{time});
    
    # Check if the mail was rejected or not
    # '.*said:\s5\d\d(?:\s|-).*'
    if (($arg_ref->{status} eq "bounced") && ($arg_ref->{reason} =~ /${$self->{regex}}[7]/x)) {
        $arg_ref->{status} = "rejected";
    }
    
    # Spezialitaet von Postfix: Host unreachable liefert "relay=none". Host muss aus SMTP-Error ausgeschnittenwerden
    # connect\s+to\s+(.*\[.*\]):.*
    if ($arg_ref->{relayhost} eq "none") {
        # 'connect to (.*)\[.*\].*'
        $arg_ref->{reason} =~ /${$self->{regex}}[9]/x;
        $arg_ref->{relayhost} = lc($1);
    }
    
    # Save the extracted values to database
    # If not explicitly set to ignore save it
    if (lc(${$self->{configuration}}{general}{ignore_verify_log_entries} ne 'yes')) {
        $self->save_deliver_entry({
            table_timestamp     => $table_timestamp,
            time                => $arg_ref->{time},
            server              => $arg_ref->{server},
            queueid             => $arg_ref->{queueid},
            mailto_address      => $arg_ref->{mailto_address},
            mailto_domain       => $arg_ref->{mailto_domain},
            original_to_address => $arg_ref->{original_to_address},
            original_to_domain  => $arg_ref->{original_to_domain},
            relayhost           => $arg_ref->{relayhost},
            status              => $arg_ref->{status},
            reason              => $arg_ref->{reason}
        });
    } else {
        if (!(($arg_ref->{status} eq 'deliverable') || ($arg_ref->{status} eq 'undeliverable'))) {
            $self->save_deliver_entry({
                table_timestamp     => $table_timestamp,
                time                => $arg_ref->{time},
                server              => $arg_ref->{server},
                queueid             => $arg_ref->{queueid},
                mailto_address      => $arg_ref->{mailto_address},
                mailto_domain       => $arg_ref->{mailto_domain},
                original_to_address => $arg_ref->{original_to_address},
                original_to_domain  => $arg_ref->{original_to_domain},
                relayhost           => $arg_ref->{relayhost},
                status              => $arg_ref->{status},
                reason              => $arg_ref->{reason}
            });
        }
    }
    
    return;   
}

# Convert "month dd hh:mm:ss" to yyyy-mm-dd hh:mm:ss
sub convert_time {
	my ($self, $date) = @_;
	my $converted_time;
	my $table_timestamp;
	
	# Check which date format is to be used
	if (lc(${$self->{configuration}}{general}{date_format}) eq "classic") { 
    	my @date_parts = split(/\s+/x, $date);
    	$converted_time = $self->{year}."-".${$self->{month_table}}{$date_parts[0]}."-".sprintf('%02d', $date_parts[1])." ".$date_parts[2];
    	$table_timestamp = $self->{year}.${$self->{month_table}}{$date_parts[0]}.sprintf('%02d', $date_parts[1]).'000000';
	
	} elsif (lc(${$self->{configuration}}{general}{date_format}) eq "modern") {
	    my @date_parts = split(/T/x, $date);
	    my @time_parts = split(/\./x, $date_parts[1]);
	    
	    $converted_time = $date_parts[0]." ".$time_parts[0];
	    $date_parts[0] =~ s/-//g;
	    $table_timestamp = $date_parts[0].'000000';
	}
	
	return ($converted_time, $table_timestamp);
}

# Check if a connectlog entry is complete for saving
sub check_completeness {
	my ($self, $connectlog_collector) = @_;
	
	if ( defined $$connectlog_collector{TIME} &&
	     defined $$connectlog_collector{TABLETIMESTAMP} &&
		 defined $$connectlog_collector{CLIENT_HOSTNAME} &&
		 defined $$connectlog_collector{CLIENT_IP} &&
		 defined $$connectlog_collector{SERVER} &&
		 defined $$connectlog_collector{MAILFROM_ADDRESS} &&
		 defined $$connectlog_collector{MAILFROM_DOMAIN} &&
		 defined $$connectlog_collector{SIZE} &&
		 defined $$connectlog_collector{MESSAGEID}
	) {
		return 1;
	} else {
		return 0;
	}
}

sub save_deliver_entry {
    my ($self, $arg_ref) = @_;
    
	# Ok, we have an deliverlog entry. Do we have an submitted DBID from the caller?
	if (defined $arg_ref->{dbid}) {
		$self->{database}->set_deliverlog($arg_ref);
		
	# Try to get the DBID from memcached
	} elsif (defined ${$self->{memcached}}->get($arg_ref->{queueid})) {
		# Get the cached dataset
		my %data_set = %{${$self->{memcached}}->get($arg_ref->{queueid})};
		
		# Reset the table timestamp if necessary
		$arg_ref->{table_timestamp} = $data_set{TABLETIMESTAMP} if defined $data_set{TABLETIMESTAMP};
		
		# Save the log entry to database
		if (defined $data_set{DBID}) {
			$self->{database}->set_deliverlog({
			    table_timestamp     => $arg_ref->{table_timestamp},
			    dbid                => $data_set{DBID},
			    mailto_address      => $arg_ref->{mailto_address},
			    mailto_domain       => $arg_ref->{mailto_domain},
			    original_to_address => $arg_ref->{original_to_address},
			    original_to_domain  => $arg_ref->{original_to_domain},
			    status              => $arg_ref->{status},
			    reason              => $arg_ref->{reason},
			    relayhost           => $arg_ref->{relayhost},
			    queueid             => $arg_ref->{queueid},
			    server              => $arg_ref->{server},
			    time                => $arg_ref->{time}
			});
		}
		
	# The DBID isn't in cache, try to get it from database. If nothing is found, an -1 will be used for the dbid
	} else {
		my ($received_dbid, $received_table_timestamp) = $self->{database}->get_dbid($arg_ref->{queueid}, $arg_ref->{table_timestamp});
		
		# Save the log entry to database
		$self->{database}->set_deliverlog({
		    table_timestamp     => $received_table_timestamp,
		    dbid                => $received_dbid,
		    mailto_address      => $arg_ref->{mailto_address},
		    mailto_domain       => $arg_ref->{mailto_domain},
		    original_to_address => $arg_ref->{original_to_address},
		    original_to_domain  => $arg_ref->{original_to_domain},
		    status              => $arg_ref->{status},
		    reason              => $arg_ref->{reason},
		    relayhost           => $arg_ref->{relayhost},
		    queueid             => $arg_ref->{queueid},
		    server              => $arg_ref->{server},
		    time                => $arg_ref->{time}
		});
	}
	
	return;
}

1;

__END__
