package Mailtrace::Database::Layer;

# Includes
use strict;
use warnings;
use DBI;
use DBD::Pg;
use DateTime;
use utf8;
use Encode;

use Mailtrace::System::Logging;

# Constructor
sub new {
	my ($invocant, $basic_functions, $logging, $configuration, $memd) = @_;
	
	# Was the constructor called by an instance or an class?
	my $class = ref($invocant) || $invocant;
	
	# Instance is a hash reference
	my $self = {
		configuration         => $configuration,
		basic_functions       => $basic_functions,
        logging               => $logging,
        memcached             => $memd,
		cached_queries        => {},
        servers               => {}
	};
	
	# Make hashreference to an instance
	bless ($self, $class); 
	
	# Connect to the specified database-type
	$self->connect_db() or do { return 0 };
	
	# Setting SQL-statements
	$self->initialize_sql();
	
	# Build an initial id mapping for the servers table
	$self->get_servers();
	
    # Return the instance
    return $self
}

# Destructor
sub DESTROY {
	my ($self) = @_;
	
	# Be verbose if wished
	$self->{logging}->debug(1, "Destroying database connection");
	
	# Finish all statement handles
	foreach my $key (keys(%{$self->{cached_queries}})) {
	    $self->{cached_queries}{$key}->finish;
	}
    
	# Class specific things
	$self->{db_connection}->disconnect() if $self->{db_connection};
	
	# Check for an overridden destructor...
    $self->SUPER::DESTROY if $self->can("SUPER::DESTROY");
    
    return 1;
}

# Make an initial database connect on object generation
sub connect_db {
	my ($self) = @_;
	my $connect_string = "";
	
	# Check the completeness of the configuration
	if ((not defined ${$self->{configuration}}{database}{dbtype}) ||
	   (not defined ${$self->{configuration}}{database}{dbname}) ||
	   (not defined ${$self->{configuration}}{database}{dbhost}) ||
	   (not defined ${$self->{configuration}}{database}{dbport}) ||
	   (not defined ${$self->{configuration}}{database}{dbuser}) ||
	   (not defined ${$self->{configuration}}{database}{dbpassword})
	) {
	    $self->{basic_functions}->alert(1, "\n\rPlease configure all values (dbtype, dbname, dbhost, dbport, dbuser, dbpassword) in my configuration file!");
	}
	
	# Generate the appropriate connect strings
	my $connect_string_MySQL = "DBI:mysql:database=%s;host=%s;port=%s";
	my $connect_string_PostgreSQL = "DBI:Pg:dbname=%s;host=%s;port=%s";
	
	# Decide which database type is to be used
	if (lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") {
		$connect_string = $connect_string_MySQL;
	} elsif (lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") {
		$connect_string = $connect_string_PostgreSQL;
	} else {
		$self->{basic_functions}->alert(0, "The configured database type is not supported!");
	}
	
	# Fill the connection string
	$connect_string = sprintf(
		$connect_string, ${$self->{configuration}}{database}{dbname},
		${$self->{configuration}}{database}{dbhost},
		${$self->{configuration}}{database}{dbport}
	);
	
	# Create the database object
	$self->{db_connection} = DBI->connect (
		$connect_string,
		${$self->{configuration}}{database}{dbuser},
		${$self->{configuration}}{database}{dbpassword},
		{PrintError => 0, RaiseError => 0}
	) or do {
		$self->{logging}->log_syslog('crit', "Unable to connect the database (".$DBI::state."): ".$DBI::errstr);
		return 0;
	};
	
	# Return TRUE, if everything works fine
	return 1;
}

# Set SQL statements for database type
sub initialize_sql {
	my ($self) = @_;
	
	# No else catch needed, because of previous check in connect_db
    if (lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") {
        $self->{sql_create_connectlog} = 
            'CREATE TABLE IF NOT EXISTS `connectlog_%s` ('.
			'`dbid` int(11) NOT NULL auto_increment,'.
			'`c_client_hostname` varchar(256) NOT NULL,'.
			'`c_client_ip` varchar(256) NOT NULL,'.
			'`c_mailfrom` varchar(300) default NULL,'.
			'`c_mailfrom_address` varchar(256) default NULL,'.
			'`c_mailfrom_domain` varchar(256) default NULL,'.
			'`c_messageid` varchar(1024) default NULL,'.
			'`c_size` int(11) NOT NULL,'.
			'`c_queueid` varchar(16) default NULL,'.
			'`c_server` smallint NOT NULL,'.
			'`c_time` datetime NOT NULL,'.
			'`c_utime` TIMESTAMP ON UPDATE CURRENT_TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP ,'.
			'PRIMARY KEY  (`dbid`),'.
			'KEY `idx_c_mailfrom` (`c_mailfrom`),'.
			'KEY `idx_c_mailfrom_address` (`c_mailfrom_address`),'.
			'KEY `idx_c_mailfrom_domain` (`c_mailfrom_domain`),'.
			'KEY `idx_c_messageid` (`c_messageid`),'.
			'KEY `idx_c_queueid` (`c_queueid`),'.
			'KEY `idx_c_server` (`c_server`),'.
			'KEY `idx_c_time` (`c_time`),'.
			'KEY `idx_c_client_hostname` (`c_client_hostname`),'.
			'KEY `idx_c_client_ip` (`c_client_ip`),'.
			'KEY `idx_c_utime` (`c_utime`),'.
			'UNIQUE KEY `idx_c_unique` (`c_mailfrom`, `c_queueid`, `c_server`, `c_time`)'.		
			') ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;';
		
        $self->{sql_create_deliverlog} = 
			'CREATE TABLE IF NOT EXISTS `deliverlog_%s` ('.
			'`id` int(11) NOT NULL auto_increment,'.
			'`d_dbid` int(11) NOT NULL,'.
			'`d_mailto` varchar(512) default NULL,'.
			'`d_mailto_address` varchar(256) default NULL,'.
			'`d_mailto_domain` varchar(256) default NULL,'.
			'`d_origto` varchar(512) default NULL,'.
			'`d_origto_address` varchar(256) default NULL,'.
			'`d_origto_domain` varchar(256) default NULL,'.
			'`d_status` varchar(30) default NULL,'.
			'`d_reason` varchar(2048) default NULL,'.
			'`d_relayhost` varchar(2048) default NULL,'.
			'`d_queueid` varchar(16) NOT NULL,'.
			'`d_server` smallint NOT NULL,'.
			'`d_time` datetime NOT NULL,'.
			'`d_utime` TIMESTAMP ON UPDATE CURRENT_TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP ,'.
			'`d_md5sum` varchar(40) NOT NULL,'.
			'PRIMARY KEY  (`id`),'.
			'KEY `idx_d_dbid` (`d_dbid`),'.
			'KEY `idx_d_mailto` (`d_mailto`),'.
			'KEY `idx_d_mailto_address` (`d_mailto_address`),'.
			'KEY `idx_d_mailto_domain` (`d_mailto_domain`),'.
			'KEY `idx_d_status` (`d_status`),'.
			'KEY `idx_d_queueid` (`d_queueid`),'.
			'KEY `idx_d_server` (`d_server`),'.
			'KEY `idx_d_time` (`d_time`),'.
			'KEY `idx_d_relayhost` (`d_relayhost`(333)),'.
			'KEY `idx_d_reason` (`d_reason`(333)),'.
			'KEY `idx_d_origto` (`d_origto`),'.
			'KEY `idx_d_origto_address` (`d_origto_address`),'.
			'KEY `idx_d_origto_domain` (`d_origto_domain`),'.
			'KEY `idx_d_utime` (`d_utime`),'.
			'KEY `idx_d_md5sum` (`d_md5sum`),'.
			'UNIQUE KEY `idx_d_unique` (`d_md5sum`)'.
			') ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;';
		
        $self->{sql_create_server} = 
			'CREATE TABLE IF NOT EXISTS `servers` ('.
			'`id` smallint NOT NULL auto_increment,'.
			'`s_name` varchar(256) NOT NULL,'.
			'PRIMARY KEY  (`id`),'.
			'KEY `idx_name` (`s_name`)'.
			') ENGINE=MyISAM DEFAULT CHARSET=utf8 AUTO_INCREMENT=1 ;';
	    
	    $self->{sql_create_config_table} =
            'CREATE TABLE IF NOT EXISTS `config` ('.
            '`configuration_key` varchar(1024) NOT NULL,'.
            '`value` TEXT NOT NULL,'.
            '`time_created` TIMESTAMP DEFAULT CURRENT_TIMESTAMP,'.
            '`time_updated` TIMESTAMP NOT NULL'.
            ') ENGINE=MyISAM DEFAULT CHARSET=utf8 ;';
	    
	    $self->{sql_create_config_table_additives} = [
	       'CREATE TRIGGER update_stamp_config BEFORE INSERT ON config '.
	       'FOR EACH ROW SET NEW.time_updated = CURRENT_TIMESTAMP;',
	       'INSERT INTO config (configuration_key,value) VALUES (\'version\',\'1.0\');'
	    ];
	    
        $self->{sql_show_tables} = 'SHOW TABLES';
        
        $self->{sql_insert_deliverlog} = "INSERT INTO deliverlog_%s (d_dbid, d_mailto, d_mailto_address, d_mailto_domain, d_origto, d_origto_address, d_origto_domain, d_status, d_reason, d_relayhost, d_queueid, d_server, d_time, d_md5sum) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, MD5(?))";
		
	# Or the sql statements depending on postgresql	
	} elsif (lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") {
        $self->{sql_create_connectlog} = 
			'CREATE TABLE connectlog_%s ('.
			'dbid BIGSERIAL PRIMARY KEY,'.
			'c_client_hostname varchar(256) NOT NULL,'.
			'c_client_ip varchar(256) NOT NULL,'.
			'c_mailfrom varchar(512) default NULL,'.
			'c_mailfrom_address varchar(256) default NULL,'.
			'c_mailfrom_domain varchar(256) default NULL,'.
			'c_messageid varchar(1024) default NULL,'.
			'c_size INTEGER NOT NULL,'.
			'c_queueid varchar(16) default NULL,'.
			'c_server SMALLINT NOT NULL,'.
			'c_time TIMESTAMP NOT NULL,'.
			'c_utime TIMESTAMP WITHOUT TIME ZONE DEFAULT CURRENT_TIMESTAMP);';
		
		$self->{sql_create_connectlog_additives} = [
		    'CREATE INDEX idx_%s_c_client_hostname ON connectlog_%s USING btree(c_client_hostname);',
            'CREATE INDEX idx_%s_c_client_ip ON connectlog_%s USING btree(c_client_ip);',
            'CREATE INDEX idx_%s_c_mailfrom ON connectlog_%s USING btree(c_mailfrom);',
            'CREATE INDEX idx_%s_c_mailfrom_address ON connectlog_%s USING btree(c_mailfrom_address);',
            'CREATE INDEX idx_%s_c_mailfrom_domain ON connectlog_%s USING btree(c_mailfrom_domain);',
            'CREATE INDEX idx_%s_c_messageid ON connectlog_%s USING btree(c_messageid);',
            'CREATE INDEX idx_%s_c_size ON connectlog_%s USING btree(c_size);',
            'CREATE INDEX idx_%s_c_queueid ON connectlog_%s USING btree(c_queueid);',
            'CREATE INDEX idx_%s_c_server ON connectlog_%s USING btree(c_server);',
            'CREATE INDEX idx_%s_c_time ON connectlog_%s USING btree(c_time);',
            'CREATE INDEX idx_%s_c_utime ON connectlog_%s USING btree(c_utime);',
            'CREATE UNIQUE INDEX idx_%s_c_unique ON connectlog_%s USING btree(c_mailfrom,c_queueid,c_server,c_time);'
		];
			
        $self->{sql_create_deliverlog} = 
			'CREATE TABLE deliverlog_%s ('.
			'id BIGSERIAL PRIMARY KEY,'.
			'd_dbid INTEGER NOT NULL,'.
			'd_mailto varchar(512) default NULL,'.
			'd_mailto_address varchar(256) default NULL,'.
			'd_mailto_domain varchar(256) default NULL,'.
			'd_origto varchar(512) default NULL,'.
			'd_origto_address varchar(256) default NULL,'.
			'd_origto_domain varchar(256) default NULL,'.
			'd_status varchar(30) default NULL,'.
			'd_reason varchar(2048) default NULL,'.
			'd_relayhost varchar(2048) default NULL,'.
			'd_queueid varchar(16) NOT NULL,'.
			'd_server SMALLINT NOT NULL,'.
			'd_time TIMESTAMP NOT NULL,'.
			'd_utime TIMESTAMP WITHOUT TIME ZONE DEFAULT CURRENT_TIMESTAMP);';
		
		$self->{sql_create_deliverlog_additives} = [
		    'CREATE INDEX idx_%s_d_dbid ON deliverlog_%s USING btree(d_dbid);',
		    'CREATE INDEX idx_%s_d_mailto ON deliverlog_%s USING btree(d_mailto);',
            'CREATE INDEX idx_%s_d_mailto_address ON deliverlog_%s USING btree(d_mailto_address);',
            'CREATE INDEX idx_%s_d_mailto_domain ON deliverlog_%s USING btree(d_mailto_domain);',
            'CREATE INDEX idx_%s_d_origto ON deliverlog_%s USING btree(d_origto);',
            'CREATE INDEX idx_%s_d_origto_address ON deliverlog_%s USING btree(d_origto_address);',
            'CREATE INDEX idx_%s_d_origto_domain ON deliverlog_%s USING btree(d_origto_domain);',
            'CREATE INDEX idx_%s_d_status ON deliverlog_%s USING btree(d_status);',
            'CREATE INDEX idx_%s_d_reason ON deliverlog_%s USING btree(d_reason);',
            'CREATE INDEX idx_%s_d_relayhost ON deliverlog_%s USING btree(d_relayhost);',
            'CREATE INDEX idx_%s_d_queueid ON deliverlog_%s USING btree(d_queueid);',
            'CREATE INDEX idx_%s_d_server ON deliverlog_%s USING btree(d_server);',
            'CREATE INDEX idx_%s_d_time ON deliverlog_%s USING btree(d_time);',
            'CREATE INDEX idx_%s_d_utime ON deliverlog_%s USING btree(d_utime);',
            'CREATE UNIQUE INDEX idx_%s_d_unique ON deliverlog_%s USING btree(d_mailto,d_origto,d_queueid,d_server,d_time);'
		];
		
        $self->{sql_create_server} = 
			'CREATE TABLE servers ('.
			'id SERIAL PRIMARY KEY,'.
			's_name varchar(256) NOT NULL);';
		
		$self->{sql_create_server_additives} = [
		    'CREATE INDEX idx_s_name ON servers USING btree(s_name);',
		    'ALTER TABLE "public"."servers" ADD CONSTRAINT "uniq_server_name" UNIQUE ("s_name");'
		];
		
		$self->{sql_create_config_table} =
            'CREATE TABLE config ('.
            'configuration_key character varying(1024) NOT NULL,'.
            '"value" text,'.
            'time_created timestamp with time zone NOT NULL DEFAULT now(),'.
            'time_updated timestamp with time zone NOT NULL DEFAULT now(),'.
            'CONSTRAINT pri_config_key PRIMARY KEY (configuration_key));';
		
		$self->{sql_create_config_table_additives} = [
		    'CREATE LANGUAGE plpgsql;',
            'CREATE OR REPLACE FUNCTION set_update_time() '.
            'RETURNS trigger AS '.
            ' $BODY$ '.
            'BEGIN '.
            'NEW.time_updated := current_timestamp; '.
            'RETURN NEW; '.
            'END; '.
            ' $BODY$ '.
            'LANGUAGE plpgsql VOLATILE ',
            'CREATE TRIGGER update_stamp_config '.
            'BEFORE INSERT OR UPDATE '.
            'ON config '.
            'FOR EACH ROW '.
            'EXECUTE PROCEDURE set_update_time(); ',
            'INSERT INTO config (configuration_key,value) VALUES (\'version\',\'1.0\');'
        ];
		
        $self->{sql_show_tables} = 
			'SELECT table_name FROM information_schema.tables WHERE table_schema = \'public\'';
			
		$self->{sql_insert_deliverlog} = "INSERT INTO deliverlog_%s (d_dbid, d_mailto, d_mailto_address, d_mailto_domain, d_origto, d_origto_address, d_origto_domain, d_status, d_reason, d_relayhost, d_queueid, d_server, d_time) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
	}
	
	# Generic stuff here
	$self->{sql_delete_table}         = "DROP TABLE %s";
	$self->{sql_insert_connectlog}    = "INSERT INTO connectlog_%s (c_client_hostname, c_client_ip, c_mailfrom, c_mailfrom_address, c_mailfrom_domain, c_messageid, c_size, c_queueid, c_server, c_time) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
	$self->{sql_insert_server}        = "INSERT INTO servers (s_name) VALUES (?)";
	$self->{sql_select_server}        = "SELECT id, s_name FROM servers";
	$self->{sql_simple_select_dbid}   = "SELECT dbid FROM %s WHERE c_queueid=?";
	$self->{sql_complex_select_dbid}  = "SELECT dbid FROM %s WHERE c_mailfrom_address=? AND c_mailfrom_domain=? AND c_queueid=? AND c_server=? AND c_time=? LIMIT 1";
	$self->{sql_select_deliverlog_id} = "SELECT id FROM %s WHERE d_mailto_address=? AND d_mailto_domain=? AND d_queueid=? AND d_reason=? AND d_server=? AND d_time=? LIMIT 1";
	$self->{sql_get_configuration}    = "SELECT configuration_key, value FROM config";
	$self->{sql_select_noqueue_dbid}  = "SELECT dbid FROM %s WHERE c_mailfrom=? AND c_queueid=? AND c_server=? AND c_time=? LIMIT 1";
	
	return 1;
}

# IMPORTANT: Each cached query needs an unique identifier!
# So be careful when using this function, else strange problems occur!
sub cache_query {
    my ($self, $identifier, $query) = @_;
    
    # Prepare the query
	$self->{cached_queries}{$identifier} = $self->{db_connection}->prepare_cached($query) or do {
		# If handleError gives TRUE, try the action again
		if ($self->handle_error("cache_query", "Unable to prepare statement for: ".$identifier)) {
			$self->cache_query($identifier, $query);
			return 1;
			
		# Else skip this dataset
		} else {
			return 0;
		}
	};
    
    return 1;
}

sub read_configuration {
    my ($self) = @_;
    my %db_configuration;
    
    # Prepare the query
    $self->cache_query("read_configuration", $self->{sql_get_configuration})
        if not defined $self->{cached_queries}{"read_configuration"};
    
    # Execute the query
    $self->{cached_queries}{"read_configuration"}->execute() or do {
        # If handle_error gives TRUE, try the action again
		if ($self->handle_error('read_configuration', 'Unable to execute statement for getting configuration')) {
			%db_configuration = %{$self->read_configuration()};
			return \%db_configuration;
		} else {
			return 0;
		}
    };
    
    # Write the db-values to the hash
    while (my @result = $self->{cached_queries}{"read_configuration"}->fetchrow_array) {
		$db_configuration{$result[0]} = $result[1];
	}
    
    return \%db_configuration;
}

sub clean_tables {
	my ($self, $max_age) = @_;
	my $tables = $self->get_table_names('(?:connectlog|deliverlog)');
	
	# Calculate the deletion date
	my $deletion_start = DateTime->now->subtract(days => $max_age);
	my $deletion_start_formatted = $deletion_start->ymd;
	$deletion_start_formatted =~ s/-//gx;
	$deletion_start_formatted .= '000000';
	
	# Now go through all found tables and delete the old ones
	foreach (@$tables) {
		my $table = $_;
		
		# Quick bugfix, because old appliances (<1.0.9) have single "connectlog" & "deliverlog"
		if (($table eq 'deliverlog') || ($table eq 'connectlog')) { next; }
		
		if (substr($table, 11) < $deletion_start_formatted) {
		    # Generate cached query if not already happen
            $self->cache_query("clean_table_$table", sprintf($self->{sql_delete_table}, $table))
                if not defined $self->{cached_queries}{"clean_table_$table"};
			
			$self->{cached_queries}{"clean_table_$table"}->execute() or do {
				# If handle_error gives TRUE, try the action again
				if ($self->handle_error('clean_entries', 'Unable to execute statement for deleting old tables')) {
					$self->clean_tables($max_age);
					return 1;
				} else {
					return 0;
				}
			};
			
			# Log the work
			$self->{logging}->log_syslog('info', "Deleted old table '".$table."'");
		}
	}
	
	return 1;
}

sub create_basic_tables {
	my ($self) = @_;
	
	# Generate cached query if not already happen
    $self->cache_query("create_basic_tables", $self->{sql_create_server})
        if not defined $self->{cached_queries}{"create_basic_tables"};
	
	$self->{cached_queries}{"create_basic_tables"}->execute() or do {
		# If handleError gives TRUE, try the action again
		if ($self->handle_error('create_basic_tables', 'Unable to prepare statement for creating basic tables')) {
            $self->create_basic_tables();
            return 1;
		} else {
            return 0;
		}
	};
	
	$self->execute_additives('sql_create_server_additives') if defined $self->{sql_create_server_additives};
	
	return 1;
}

sub create_config_table {
    my ($self) = @_;
    
    # Generate cached query if not already happen
    $self->cache_query("create_config_tables", $self->{sql_create_config_table})
        if not defined $self->{cached_queries}{"create_config_tables"};
	
	$self->{cached_queries}{"create_config_tables"}->execute() or do {
		# If handleError gives TRUE, try the action again
		if ($self->handle_error('create_config_tables', 'Unable to prepare statement for creating the config table')) {
            $self->create_config_tables();
            return 1;
		} else {
            return 0;
		}
	};
	
	$self->execute_additives('sql_create_config_table_additives') if defined $self->{sql_create_config_table_additives};
    
    return 1;
}

sub create_log_tables {
	my ($self, $type, $table_timestamp) = @_;
	
	if ($type eq "connectlog") {
	    # Generate cached query if not already happen
    	$self->cache_query("create_connectlog_table_$table_timestamp", sprintf($self->{sql_create_connectlog}, $table_timestamp))
    	    if not defined $self->{cached_queries}{"create_connectlog_table_$table_timestamp"};
		
		$self->{cached_queries}{"create_connectlog_table_$table_timestamp"}->execute() or do {
			# If handleError gives TRUE, try the action again
			if ($self->handle_error('create_connectlog', 'Unable to prepare statement for creating new connectlog table', {tabletimestamp => $table_timestamp})) {
				$self->create_log_tables('connectlog', $table_timestamp);
				return 1;
			} else {
				return 0;
			}
		};
		
		$self->execute_additives(
		    'sql_create_connectlog_additives',
		    $table_timestamp
		) if defined $self->{sql_create_connectlog_additives};
		
	} elsif ($type eq "deliverlog") {
	    # Generate cached query if not already happen
    	$self->cache_query("create_deliverlog_table_$table_timestamp", sprintf($self->{sql_create_deliverlog}, $table_timestamp))
    	    if not defined $self->{cached_queries}{"create_deliverlog_table_$table_timestamp"};
		
		$self->{cached_queries}{"create_deliverlog_table_$table_timestamp"}->execute() or do {
			# If handleError gives TRUE, try the action again
			if ($self->handle_error('create_deliverlog', 'Unable to prepare statement for creating new deliverlog table', {tabletimestamp => $table_timestamp})) {
				$self->create_log_tables('deliverlog', $table_timestamp);
				return 1;
			} else {
				return 0;
			}
		};
		
		$self->execute_additives(
		    'sql_create_deliverlog_additives',
		    $table_timestamp
		) if defined $self->{sql_create_deliverlog_additives};
	}
	
	return 1;
}

# Execute defined additives (ATM there is only index creation for postgresql)
sub execute_additives {
    my ($self, $identifier, $table_timestamp) = @_;
    
    # Additional parameters are table_timestamp atm
    # So if nothing is submitted set NONE as default
    $table_timestamp = 'NONE' if not defined $table_timestamp;
    
    # Additives are alway arrays so loop through
    for (my $i=0; $i<@{$self->{$identifier}}; $i++) {
        # Generate cached query if not already happen
    	$self->cache_query($identifier.'_'.$table_timestamp.'_'.$i, sprintf(${$self->{$identifier}}[$i], $table_timestamp, $table_timestamp))
    	    if not defined $self->{cached_queries}{$identifier.'_'.$table_timestamp.'_'.$i};
    	
    	$self->{cached_queries}{$identifier.'_'.$table_timestamp.'_'.$i}->execute() or do {
			# If handleError gives TRUE, try the action again
			if ($self->handle_error('execute_additives', 'Unable to prepare statement for additive statement', {tabletimestamp => $table_timestamp})) {
				$self->execute_additives($identifier, $table_timestamp);
				return 1;
			} else {
				return 0;
			}
		};
    }
    
    return 1;
}

# Save an connectlog dataset to database
sub set_connectlog {
    my ($self, $arg_ref) = @_;
    
	# Check if the id from the given server is already saved
	if (not defined $self->{servers}{$arg_ref->{server}}) { $self->set_server($arg_ref->{server}); }
	
	# Generate cached query if not already happen
	$self->cache_query("insert_connectlog_$arg_ref->{table_timestamp}", sprintf($self->{sql_insert_connectlog}, $arg_ref->{table_timestamp}))
	    if not defined $self->{cached_queries}{"insert_connectlog_$arg_ref->{table_timestamp}"};
	
	# Convert the mailfrom_adress to prevent something like this: lópez025
	my $utf8_test_mailfrom_address = $arg_ref->{mailfrom_address};
	if (not utf8::decode($utf8_test_mailfrom_address)) {
        $arg_ref->{mailfrom_address} = Encode::encode("utf8", $arg_ref->{mailfrom_address}, 1);
    }
	
	# Now execute the SQL statement
	$self->{cached_queries}{"insert_connectlog_$arg_ref->{table_timestamp}"}->execute(
	    $arg_ref->{client_hostname},
	    $arg_ref->{client_ip},
	    $arg_ref->{mailfrom_address}.'@'.$arg_ref->{mailfrom_domain},
	    $arg_ref->{mailfrom_address},
	    $arg_ref->{mailfrom_domain},
	    $arg_ref->{messageid},
	    $arg_ref->{size},
	    $arg_ref->{queueid},
	    $self->{servers}{$arg_ref->{server}},
	    $arg_ref->{time}
	) or do {
		# If handleError gives TRUE, try the action again
		my $return_set = $self->handle_error("set_connectlog", "Unable to execute statement for setting connectlog entry", $arg_ref);
		
		# Try if we had an unique problem
		if ($return_set =~ /DBID-(\d*)/) {
		    # Just give back the determined id
			return $1;
		
		# Else check the success
		} elsif ($return_set == 1) {
			$self->set_connectlog($arg_ref);
			return 1;	
			
		# Else skip this dataset
		} else {
			return 0;
		}
	};
	
	# Update the backlog monitoring
	${$self->{memcached}}->set('last_processed_time', $arg_ref->{time});
	
	# Make a keep alive ping
	${$self->{memcached}}->set('keep_alive', '1', ${$self->{configuration}}{general}{keep_alive_seconds});
	
	# Return TRUE, if everything works fine
	return $self->{db_connection}->last_insert_id(undef, undef, "connectlog_$arg_ref->{table_timestamp}", "dbid") or do {
		$self->{logging}->debug(0, "Unable to get the last inserted Connectlog-ID (".$DBI::state."): ".$DBI::errstr);
	}
}

# Save an deliverlog dataset to database
sub set_deliverlog {
	my ($self, $arg_ref) = @_;
    my $mailto_address = $arg_ref->{mailto_address};
    my $original_to_address = $arg_ref->{original_to_address};
    my $original_to = '';
    
	# Check if the id from the given server is already saved
	if (not defined $self->{servers}{$arg_ref->{server}}) { $self->set_server($arg_ref->{server}); }
	
	# Generate cached query if not already happen
	$self->cache_query("insert_deliverlog_$arg_ref->{table_timestamp}", sprintf($self->{sql_insert_deliverlog}, $arg_ref->{table_timestamp}))
	    if not defined $self->{cached_queries}{"insert_deliverlog_$arg_ref->{table_timestamp}"};
	
	# Convert the mailfrom_adress to prevent something like this: lópez025
	my $utf8_test_mailto_address = $mailto_address;
	my $utf8_test_original_to_address = $original_to_address;
	if ((not utf8::decode($utf8_test_mailto_address)) || (not utf8::decode($utf8_test_original_to_address))) {
        $mailto_address = Encode::encode("utf8", $mailto_address, 1);
        $original_to_address = Encode::encode("utf8", $original_to_address, 1);
    }
    
    # Only compose, if we have an origto address
    if ((defined $original_to_address) && (defined $arg_ref->{original_to_domain})) {
	   $original_to = $original_to_address.'@'.$arg_ref->{original_to_domain};
    }
    
	# Check which DB is used
	if (lc(${$self->{configuration}}{database}{dbtype}) eq 'mysql') {        
        # Now execute the SQL statement
    	$self->{cached_queries}{"insert_deliverlog_$arg_ref->{table_timestamp}"}->execute(
            $arg_ref->{dbid},
            $mailto_address.'@'.$arg_ref->{mailto_domain},
            $mailto_address,
            $arg_ref->{mailto_domain},
            $original_to,
            $original_to_address,
            $arg_ref->{original_to_domain},
            $arg_ref->{status},
            $arg_ref->{reason},
            $arg_ref->{relayhost},
            $arg_ref->{queueid},
            $self->{servers}{$arg_ref->{server}},
            $arg_ref->{time},
            # Composed string for md5sum field
            $mailto_address.'@'.$arg_ref->{mailto_domain}.$original_to.$arg_ref->{queueid}.
            $self->{servers}{$arg_ref->{server}}.$arg_ref->{time}
        ) or do {
            if ($self->handle_error('set_deliverlog', 'Unable to execute statement for setting deliverlog entry', $arg_ref)) {
                $self->set_deliverlog($arg_ref);
                return 1;
                # Else skip this dataset
            } else {
                return 0;
            }
        };
        
	# Only "postgresql" is left
	} else {
	    # Now execute the SQL statement
    	$self->{cached_queries}{"insert_deliverlog_$arg_ref->{table_timestamp}"}->execute(
            $arg_ref->{dbid},
            $mailto_address.'@'.$arg_ref->{mailto_domain},
            $mailto_address,
            $arg_ref->{mailto_domain},
            $original_to,
            $original_to_address,
            $arg_ref->{original_to_domain},
            $arg_ref->{status},
            $arg_ref->{reason},
            $arg_ref->{relayhost},
            $arg_ref->{queueid},
            $self->{servers}{$arg_ref->{server}},
            $arg_ref->{time}
        ) or do {
            if ($self->handle_error('set_deliverlog', 'Unable to execute statement for setting deliverlog entry', $arg_ref)) {
                $self->set_deliverlog($arg_ref);
                return 1;
                # Else skip this dataset
            } else {
                return 0;
            }
        };
	}
	
	# Update the backlog monitoring
	${$self->{memcached}}->set('last_processed_time', $arg_ref->{time});
	
	# Make a keep alive ping
	${$self->{memcached}}->set('keep_alive', '1', ${$self->{configuration}}{general}{keep_alive_seconds});
	
	# Return TRUE, if everything works fine
	return 1;
}

# Update a new server entry in hash and DB
sub set_server {
	my ($self, $server) = @_;
	
	# Generate cached query if not already happen
	$self->cache_query("insert_server", $self->{sql_insert_server})
	    if not defined $self->{cached_queries}{"insert_server"};
	
	# Now execute the SQL statement
	$self->{cached_queries}{"insert_server"}->execute($server) or do {
		if ($self->handle_error('set_server', "Unable to execute statement for setting a server name")) {
			$self->set_server($server);
			return 1;
			
		# Else skip this dataset
		} else {
			return 0;
		}
	};
	
	# Save the ID and server name to the hash
	$self->{servers}{$server} = $self->{db_connection}->last_insert_id(undef, undef, "servers", "ID") or do {
		$self->{logging}->debug(0, "Unable to get the last inserted Server-ID (".$DBI::state."): ".$DBI::errstr);
	};
	
	return 1;
}

# Build an hash of saved servers an return its hash reference
sub get_servers {
	my ($self) = @_;
	
	# Generate cached query if not already happen
	$self->cache_query("select_servers", $self->{sql_select_server})
	    if not defined $self->{cached_queries}{"select_servers"};
	
	# Now execute the SQL statement
	$self->{cached_queries}{"select_servers"}->execute() or do {
		if ($self->handle_error('get_servers', "Unable to execute statement for getting the saved server names")) {
			$self->get_servers();
			return 1;
			
		# Else skip this dataset
		} else { return 0; }
	};
	
	# Get the results
	while (my @result = $self->{cached_queries}{"select_servers"}->fetchrow_array) {
		$self->{servers}{$result[1]} = $result[0];
	}
	
	# If everything went fine
	return 1;
}

sub get_table_names {
	my ($self, $type) = @_;
	my @tables;
	
	# Generate cached query if not already happen
	$self->cache_query("show_tables", $self->{sql_show_tables})
	    if not defined $self->{cached_queries}{"show_tables"};
	
	$self->{cached_queries}{"show_tables"}->execute() or do {
		if ($self->handle_error('get_table_names', "Unable to prepare statement for getting tablenames")) {
			return $self->get_table_names($type);
				
		# Else skip this dataset
		} else {
			return 0;
		}
	};
	
	while (my @result = $self->{cached_queries}{"show_tables"}->fetchrow_array) {
        if ($result[0] =~ /$type/x) {
			push(@tables, $result[0]);
		}
	}
	
	# Sort in reversed order
	@tables = sort {$b cmp $a} @tables;
	
	return \@tables;
}

# Get the DBID for an given queueid
sub get_dbid {
	my ($self, $queueid, $table_timestamp) = @_;
	my $dbid = "-1"; # Default value
	
	# Get all existing connectlog tablenames
	my $connectlog_tables = $self->get_table_names('connectlog');
	
	# Go through all found tables
	for (my $i=0; $i<@$connectlog_tables; $i++) {
	    # Generate cached query if not already happen
    	$self->cache_query("select_dbid_$$connectlog_tables[$i]", sprintf($self->{sql_simple_select_dbid}, $$connectlog_tables[$i]))
    	    if not defined $self->{cached_queries}{"select_dbid_$$connectlog_tables[$i]"};
		
		# Now execute the SQL statement
		$self->{cached_queries}{"select_dbid_$$connectlog_tables[$i]"}->execute($queueid) or do {
			if ($self->handle_error('get_dbid', "Unable to execute statement for obtaining QueueID")) {
				return $self->get_dbid($queueid);
				
			# Else skip this dataset
			} else { return 0; }
		};
		
		# Get the result
		while (my @result = $self->{cached_queries}{"select_dbid_$$connectlog_tables[$i]"}->fetchrow_array) { $dbid = $result[0]; }
		
		# Check if we got something and exit the loop then
		if ($dbid != -1) {
			$table_timestamp = substr($$connectlog_tables[$i], 11);
			last;
		}
	}
	
	# Send the obtained ID back (default is -1 and the submitted table_timestamp)
	return ($dbid, $table_timestamp);
}

sub get_noqueue_dbid {
    my ($self, $table_timestamp, $mailfrom, $queueid, $server, $time) = @_;
    my $dbid = "-2"; # Default value
    
    # Just check the current table, because the entry has to be there, regarding to the timing
    $self->cache_query("get_noqueue_dbid_$table_timestamp", sprintf($self->{sql_select_noqueue_dbid}, "connectlog_$table_timestamp"))
    if not defined $self->{cached_queries}{"get_noqueue_dbid_$table_timestamp"};

	# Now execute the SQL statement
    $self->{cached_queries}{"get_noqueue_dbid_$table_timestamp"}->execute($mailfrom, $queueid, $self->{servers}{$server},, $time) or do {
		if ($self->handle_error('get_noqueue_dbid', "Unable to execute statement for obtaining DBID for an NOQUEUE entry")) {
			return $self->get_dbid($table_timestamp, $mailfrom, $queueid, $server, $time);
			
		# Else skip this dataset
		} else { return 0; }
	};
	
	# Get the result
	while (my @result = $self->{cached_queries}{"get_noqueue_dbid_$table_timestamp"}->fetchrow_array) { $dbid = $result[0]; }
	
	# Give back the answer
	return $dbid;
}

# Handle errors occured during DB operations (except the connection phase)
sub handle_error {
	my ($self, $control, $msg, $arg_ref) = @_;
	
	# Catch MySQL and PostgreSQL errors
	# MySQL: 	  42S02 Table doesn't exist
	# PostgreSQL: 42P01 Table doesn't exist
	# MySQL:      23000 Duplicate entry
	# PostgreSQL: 23505 unique violation
	# MySQL:      1021  Disk full
	# PostgreSQL: 53100 No space left
	#
	# Catch MySQL and PostgreSQL database connection losses
	# MySQL: 2006 Server has gone away
	# PostgreSQL: 08000 Connection exception
	# PostgreSQL: 57P01 Admin Shutdown
	
	if (((lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") && ($DBI::state eq '42S02')) ||
		((lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") && ($DBI::state eq '42P01'))) {
		$self->{logging}->debug(1, "Table didn't exist yet. Creating a new one...");
		$self->{logging}->debug(1, "ORIGINAL ERROR: ".$msg." (".$DBI::state."): ".$DBI::errstr);
		
		if ($control eq 'set_connectlog') {
			$self->create_log_tables('connectlog', $arg_ref->{table_timestamp});
		} elsif ($control eq 'set_deliverlog') {
			$self->create_log_tables('deliverlog', $arg_ref->{table_timestamp});
		} elsif ($control eq 'get_servers') {
			$self->create_basic_tables();
		} elsif ($control eq 'read_configuration') {
		    $self->create_config_table();
		}
		
	} elsif (((lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") && ($DBI::err == 2006)) ||
	    ((lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") && (($DBI::state eq "08000") || ($DBI::state eq "57P01"))) ||
	    ((lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") && ($DBI::err eq "1021")) ||
	    ((lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") && ($DBI::state eq "53100"))) {
	    # Print info to syslog
	    $self->{logging}->log_syslog('crit', $msg." (".$DBI::state."): ".$DBI::errstr);
	    
	    # Log regarding to the problem, starting with the connection loss
	    if (($DBI::err == 2006) || ($DBI::state eq "08000")) {
            $self->{logging}->log_syslog('crit', "Going into database reconnection loop ...");            
        # Or disk full
	    } elsif (($DBI::err eq "1021") || ($DBI::state eq "53100")) {
	         $self->{logging}->log_syslog('crit', "Waiting, because disk space of database host is full ...");
	    }
	    
		# Try to reconnect the database
		while (1) {
			# Try to connect the database and quit loop if successful
			if (($DBI::err == 2006) || ($DBI::state eq "08000")) {
    			if ($self->connect_db) {
    				$self->{logging}->debug(1, "Destroying all cached statement handles for the broken connection");
    				# Delete our caching, because it refers to the old broken connection
    				foreach my $key (keys(%{$self->{cached_queries}})) {
    	               $self->{cached_queries}{$key}->finish;
    	               delete $self->{cached_queries}{$key};
    	            }
    	            $self->{logging}->log_syslog('crit', "Database connection reestablished. Continuing my work...");
    				last;
    			}
    			
    		# If the db disk is full
			} elsif (($DBI::err eq "1021") || ($DBI::state eq "53100")) {
			    # Just wait the configured time
			    if (defined ${$self->{configuration}}{database}{db_disk_full_sleeptime_seconds}) {
				    sleep ${$self->{configuration}}{database}{db_disk_full_sleeptime_seconds}
                # Or sleep the default amount of time (1800s)
                } else {
				    sleep 1800;
                } 
                # Quit the loop
                last;
			}
			
			# If not successful sleep the configured amount of time
			if (defined ${$self->{configuration}}{database}{db_disconnect_sleeptime_seconds}) {
				sleep ${$self->{configuration}}{database}{db_disconnect_sleeptime_seconds}
			# Or sleep the default amount of time (15s)
			} else {
				sleep 15;
			}
		}
		
	# Handle duplicate NOQUEUE warnings
    } elsif (($control eq 'set_connectlog') && (
        ($DBI::state eq "23000") && (lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") ||
        ($DBI::state eq "23505") && (lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql"))) {
            
        # Try to get an DBID
        my $DBID = $self->get_noqueue_dbid(
            $arg_ref->{table_timestamp},
            $arg_ref->{mailfrom_address}.'@'.$arg_ref->{mailfrom_domain},
            $arg_ref->{queueid},
            $arg_ref->{server},
            $arg_ref->{time}
        );
        
        # Look if we have something
        if (defined $DBID) {
            # Return the marked DBID
            return "DBID-".$DBID;
            
        # Else, just log
        } else {
            $self->{logging}->debug(1, "Unique violation in database");
            $self->{logging}->debug(1, "ORIGINAL ERROR: ".$msg." (".$DBI::state."): ".$DBI::errstr);
            return 0;
        }
            
	# Print out unique violations only in debug mode
	} elsif (((lc(${$self->{configuration}}{database}{dbtype}) eq "mysql") && ($DBI::state eq '23000')) ||
        ((lc(${$self->{configuration}}{database}{dbtype}) eq "postgresql") && ($DBI::state eq '23505'))) {
		$self->{logging}->debug(1, "Unique violation in database");
		$self->{logging}->debug(1, "ORIGINAL ERROR: ".$msg." (".$DBI::state."): ".$DBI::errstr);
		return 0;
        
        
	# If i can't find my configuration table
	} elsif (($control eq 'read_configuration') && ($DBI::state eq "xxxxx")) {
		$self->{logging}->log_syslog('crit', 'The configuration table does\'nt exist. I will defaulting the values as far as possible!');
		return 0;
		
	# Log any other failure and continue
	} else {
	    $self->{logging}->log_syslog('crit', $msg." (".$DBI::state."): ".$DBI::errstr);
	    return 0;
	}
	
	# If returning 1 be shure, that everything is fine, because a new try will almost be initiated
	return 1;
}

1;

__END__